﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter1
{
	public class Point
	{
		public int X
		{
			get;
			set;
		}
		public int Y
		{
			get;
			set;
		}
		public double Distance => Math.Sqrt(X * X + Y * Y);
	}

	public class Sample
	{
		public string LastName
		{
			// getter elided
			set
			{
				if (string.IsNullOrEmpty(value))
					throw new ArgumentException(
					"last name can't be null or blank");
				lastName = value;
			}
		}

		private string lastName;

	}

	namespace Sample2
	{
		public class Point
		{
			private int xValue;
			public int X
			{
				get => xValue;
				set
				{
					xValue = value;
					distance = default(double?);
				}
			}
			private int yValue;
			public int Y
			{
				get => yValue;
				set
				{
					yValue = value;
					distance = default(double?);
				}
			}
			private double? distance;
			public double Distance
			{
				get
				{
					if (!distance.HasValue)
						distance = Math.Sqrt(X * X + Y * Y);
					return distance.Value;
				}
			}
		}
	}

	namespace Sample3
	{
		// Bad Property Design. Lengthy operation required for getter
		public class MyType
		{
			// lots elided
			public string ObjectName =>
				RetrieveNameFromRemoteDatabase();

			private String RetrieveNameFromRemoteDatabase()
			{
				throw new NotImplementedException();
			}
		}
	}

	namespace Sample4
	{
		// One possible path: evaluate once and cache the answer
		public class MyType
		{
			// lots elided
			private string objectName;
			public string ObjectName =>
				(objectName != null) ?
				objectName : RetrieveNameFromRemoteDatabase();

            private Lazy<string> lazyObjectName;
            public MyType()
            {
                lazyObjectName = new Lazy<string>(() => RetrieveNameFromRemoteDatabase());
            }
            //public string ObjectName => lazyObjectName.Value;

			private String RetrieveNameFromRemoteDatabase()
			{
				throw new NotImplementedException();
			}
		}
	}

	namespace Sample5
	{
		//Better solution: Use methods to manage cached values
		public class MyType
		{
			public void LoadFromDatabase()
			{
				ObjectName = RetrieveNameFromRemoteDatabase();
				// other fields elided.
			}

			public void SaveToDatabase()
			{
				SaveNameToRemoteDatabase(ObjectName);
				// other fields elided.
			}

			// lots elided
			public string ObjectName
			{
				get;
				set;
			}


			private String RetrieveNameFromRemoteDatabase()
			{
				throw new NotImplementedException();
			}
			private void SaveNameToRemoteDatabase(String objectName)
			{
				throw new NotImplementedException();
			}
		}
	}


	namespace Sample6
	{
		// Poor property design: setter is too expensive
		public class MyType
		{
			// lots elided
			private string objectName;
			public string ObjectName
			{
				get
				{
					if (objectName == null)
						objectName = RetrieveNameFromRemoteDatabase();
					return objectName;
				}
				set
				{
					objectName = value;
					SaveNameToRemoteDatabase(objectName);
				}
			}

			private void SaveNameToRemoteDatabase(String objectName)
			{
				throw new NotImplementedException();
			}

			private String RetrieveNameFromRemoteDatabase()
			{
				throw new NotImplementedException();
			}
		}
	}
	static class Item06
	{
		private class Thing
		{
			public Int32 ImportantProperty { get; internal set; }
		}
		private static Thing someObject;

		public static void Tests()
		{
			var myArray = new int[5];

			int someValue = someObject.ImportantProperty;
			Debug.Assert(someValue == someObject.ImportantProperty);

			for (int index = 0; index < myArray.Length; index++)
				;
		}
	}
}
