﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter2
{
    public class LoggerEventArgs : EventArgs
    {
        public string Message { get; }
        public int Priority { get; }

        public LoggerEventArgs(int p, string m)
        {
            Priority = p;
            Message = m;
        }
    }

    public class Logger
    {
        static Logger()
        {
            Singleton = new Logger();
        }

        private Logger()
        {
        }

        public static Logger Singleton { get;  }

        // Define the event:
        public event EventHandler<LoggerEventArgs> Log;

        // add a message, and log it.
        public void AddMsg(int priority, string msg) =>
            Log?.Invoke(this, new LoggerEventArgs(priority, msg));
    }
    class ConsoleLogger
    {
        static ConsoleLogger() =>
            Logger.Singleton.Log += (sender, msg) =>
                Console.Error.WriteLine("{0}:\t{1}",
                    msg.Priority.ToString(),
                    msg.Message);
    }
    class EventLogger
    {
        private static Logger logger = Logger.Singleton;
        private static string eventSource;
        private static EventLog logDest;

        static EventLogger() =>
            logger.Log += (sender, msg) =>
            {
                logDest?.WriteEntry(msg.Message,
                    EventLogEntryType.Information,
                    msg.Priority);
            };

        public static string EventSource
        {
            get { return eventSource; }

            set
            {
                eventSource = value;
                if (!EventLog.SourceExists(eventSource))
                    EventLog.CreateEventSource(eventSource,
                      "ApplicationEventLogger");

                logDest?.Dispose();
                logDest = new EventLog();
                logDest.Source = eventSource;
            }
        }
    }

    public class Item16
    {
        public static void Tests()
        {

        }
    }

    namespace SampleTwo
    {
        public class Logger
        {
            private EventHandler<LoggerEventArgs> log;

            public event EventHandler<LoggerEventArgs> Log
            {
                add { log = log + value; }
                remove { log = log - value; }
            }

            public void AddMsg(int priority, string msg) =>
                log?.Invoke(this, new LoggerEventArgs(priority, msg));
        }
    }

    namespace SampleThree
    { 
        public sealed class Logger
        {
            private static EventHandlerList
              Handlers = new EventHandlerList();

            static public void AddLogger(
              string system, EventHandler<LoggerEventArgs> ev) =>
                Handlers.AddHandler(system, ev);

            static public void RemoveLogger(string system,
                EventHandler<LoggerEventArgs> ev) =>
                Handlers.RemoveHandler(system, ev);

            static public void AddMsg(string system,
              int priority, string msg)
            {
                if (!string.IsNullOrEmpty(system))
                {
                    EventHandler<LoggerEventArgs> handler =
                      Handlers[system] as
                      EventHandler<LoggerEventArgs>;

                    LoggerEventArgs args = new LoggerEventArgs(
                      priority, msg);
                    if (handler != null)
                        handler(null, args);

                    // The empty string means receive all messages:
                    handler = Handlers[""] as
                        EventHandler<LoggerEventArgs>;
                    if (handler != null)
                        handler(null, args);
                }
            }
        }
    }

    namespace Sample5
    {
        public sealed class Logger
        {
            private static Dictionary<string,
                EventHandler<LoggerEventArgs>>
              Handlers = new Dictionary<string,
                  EventHandler<LoggerEventArgs>>();

            static public void AddLogger(
              string system, EventHandler<LoggerEventArgs> ev)
            {
                if (Handlers.ContainsKey(system))
                    Handlers[system] += ev;
                else
                    Handlers.Add(system, ev);
            }

            // will throw exception if system
            // does not contain a handler.
            static public void RemoveLogger(string system,
                EventHandler<LoggerEventArgs> ev) =>
                Handlers[system] -= ev;

            static public void AddMsg(string system,
              int priority, string msg)
            {
                if (string.IsNullOrEmpty(system))
                {
                    EventHandler<LoggerEventArgs> handler = null;
                    Handlers.TryGetValue(system, out handler);

                    LoggerEventArgs args = new LoggerEventArgs(
                      priority, msg);
                    handler?.Invoke(null, args);

                    // The empty string means receive all messages:
                    handler = Handlers[""] as
                        EventHandler<LoggerEventArgs>;
                    handler?.Invoke(null, args);
                }
            }
        }

    }
}
